searchNodes=[{"type":"module","doc":"The Vettore library is designed for fast, in-memory operations on vector (embedding) data. All vectors (embeddings) are stored in a Rust data structure (a HashMap ), accessed via a shared resource (using Rustler’s ResourceArc with a Mutex ). Core operations include: - Creating a collection : A named set of embeddings with a fixed dimension and a chosen similarity metric ( :cosine , :euclidean , :dot , :hnsw , :binary ) . - Inserting an embedding : Add a new embedding ( with ID , vector , and optional metadata ) to a specific collection . - Retrieving embeddings : Fetch all embeddings from a collection or look up a single embedding by its unique ID . - Similarity search : Given a query vector , calculate a “ score ” for every embedding in the collection and return the top ‑ k results ( e . g . the smallest distances or largest similarities ) . # Usage Example db = Vettore . new ( ) :ok = Vettore . create_collection ( db , &quot;my_collection&quot; , 3 , :euclidean ) # Insert an embedding via struct: embedding = % Vettore.Embedding { value : &quot;my_id or text&quot; , vector : [ 1.0 , 2.0 , 3.0 ] , metadata : %{ &quot;note&quot; =&gt; &quot;hello&quot; } } :ok = Vettore . insert ( db , &quot;my_collection&quot; , embedding ) # Retrieve it back: { :ok , returned_emb } = Vettore . get_by_value ( db , &quot;my_collection&quot; , &quot;my_id&quot; ) IO . inspect ( returned_emb . vector , label : &quot;Retrieved vector&quot; ) # Perform a similarity search: { :ok , top_results } = Vettore . similarity_search ( db , &quot;my_collection&quot; , [ 1.5 , 1.5 , 1.5 ] , 2 ) IO . inspect ( top_results , label : &quot;Top K search results&quot; )","title":"Vettore","ref":"Vettore.html"},{"type":"function","doc":"Batch‑insert a list of embeddings. Reject elements that are not %Vettore.Embedding{} . Batch is faster than insert/3 for a large number of embeddings as it avoids to validate vector list on each embedding. Examples iex&gt; Vettore . new ( ) |&gt; Vettore . create_collection ( &quot;my_collection&quot; , 3 , :euclidean ) |&gt; Vettore . batch ( &quot;my_collection&quot; , [ % Vettore.Embedding { value : &quot;my_id&quot; , vector : [ 1.0 , 2.0 , 3.0 ] , metadata : %{ &quot;note&quot; =&gt; &quot;hello&quot; } } ] ) { :ok , [ &quot;my_id&quot; ] }","title":"Vettore.batch/3","ref":"Vettore.html#batch/3"},{"type":"function","doc":"Batch‑insert a list of embeddings. Reject elements that are not %Vettore.Embedding{} . Raise error if the metadata is invalid. Otherwise, return the list of inserted as in batch/3 . Examples iex&gt; Vettore . new ( ) |&gt; Vettore . create_collection ( &quot;my_collection&quot; , 3 , :euclidean ) |&gt; Vettore . batch ( &quot;my_collection&quot; , [ % Vettore.Embedding { value : &quot;my_id&quot; , vector : [ 1.0 , 2.0 , 3.0 ] , metadata : %{ &quot;note&quot; =&gt; [ 1 , 2 , 4 ] } } ] ) ArgumentError [ … ]","title":"Vettore.batch!/3","ref":"Vettore.html#batch!/3"},{"type":"function","doc":"Create a collection. distance must be one of the atoms: :euclidean , :cosine , :dot , :hnsw , or :binary . Examples iex&gt; Vettore . new ( ) |&gt; Vettore . create_collection ( &quot;my_collection&quot; , 3 , :euclidean ) { :ok , &quot;my_collection&quot; }","title":"Vettore.create_collection/5","ref":"Vettore.html#create_collection/5"},{"type":"function","doc":"Delete a single embedding. Examples iex&gt; Vettore . new ( ) |&gt; Vettore . create_collection ( &quot;my_collection&quot; , 3 , :euclidean ) |&gt; Vettore . insert ( &quot;my_collection&quot; , % Vettore.Embedding { value : &quot;my_id&quot; , vector : [ 1.0 , 2.0 , 3.0 ] , metadata : %{ &quot;note&quot; =&gt; &quot;hello&quot; } } ) |&gt; Vettore . delete ( &quot;my_collection&quot; , &quot;my_id&quot; ) { :ok , &quot;my_id&quot; }","title":"Vettore.delete/3","ref":"Vettore.html#delete/3"},{"type":"function","doc":"Delete a collection. Examples iex&gt; Vettore . new ( ) |&gt; Vettore . create_collection ( &quot;my_collection&quot; , 3 , :euclidean ) |&gt; Vettore . delete_collection ( &quot;my_collection&quot; ) { :ok , &quot;my_collection&quot; }","title":"Vettore.delete_collection/2","ref":"Vettore.html#delete_collection/2"},{"type":"function","doc":"Return all embeddings in raw form ( {value, vector, metadata} tuples). Examples iex&gt; Vettore . new ( ) |&gt; Vettore . create_collection ( &quot;my_collection&quot; , 3 , :euclidean ) |&gt; Vettore . insert ( &quot;my_collection&quot; , % Vettore.Embedding { value : &quot;my_id&quot; , vector : [ 1.0 , 2.0 , 3.0 ] , metadata : %{ &quot;note&quot; =&gt; &quot;hello&quot; } } ) |&gt; Vettore . get_all ( &quot;my_collection&quot; ) { :ok , [ { &quot;my_id&quot; , [ 1.0 , 2.0 , 3.0 ] , %{ &quot;note&quot; =&gt; &quot;hello&quot; } } ] }","title":"Vettore.get_all/2","ref":"Vettore.html#get_all/2"},{"type":"function","doc":"Fetch a single embedding by value (ID) and return it as %Vettore.Embedding{} . Examples iex&gt; Vettore . new ( ) |&gt; Vettore . create_collection ( &quot;my_collection&quot; , 3 , :euclidean ) |&gt; Vettore . insert ( &quot;my_collection&quot; , % Vettore.Embedding { value : &quot;my_id&quot; , vector : [ 1.0 , 2.0 , 3.0 ] , metadata : %{ &quot;note&quot; =&gt; &quot;hello&quot; } } ) |&gt; Vettore . get_by_value ( &quot;my_collection&quot; , &quot;my_id&quot; ) { :ok , % Vettore.Embedding { value : &quot;my_id&quot; , vector : [ 1.0 , 2.0 , 3.0 ] , metadata : %{ &quot;note&quot; =&gt; &quot;hello&quot; } } }","title":"Vettore.get_by_value/3","ref":"Vettore.html#get_by_value/3"},{"type":"function","doc":"Fetch a single embedding by vector and return it as %Vettore.Embedding{} . Examples iex&gt; Vettore . new ( ) |&gt; Vettore . create_collection ( &quot;my_collection&quot; , 3 , :euclidean ) |&gt; Vettore . insert ( &quot;my_collection&quot; , % Vettore.Embedding { value : &quot;my_id&quot; , vector : [ 1.0 , 2.0 , 3.0 ] , metadata : %{ &quot;note&quot; =&gt; &quot;hello&quot; } } ) |&gt; Vettore . get_by_vector ( &quot;my_collection&quot; , [ 1.0 , 2.0 , 3.0 ] ) { :ok , % Vettore.Embedding { value : &quot;my_id&quot; , vector : [ 1.0 , 2.0 , 3.0 ] , metadata : %{ &quot;note&quot; =&gt; &quot;hello&quot; } } }","title":"Vettore.get_by_vector/3","ref":"Vettore.html#get_by_vector/3"},{"type":"function","doc":"Insert one %Vettore.Embedding{} into the collection. Returns {:ok, value} on success or {:error, reason} . Examples iex&gt; Vettore . new ( ) |&gt; Vettore . create_collection ( &quot;my_collection&quot; , 3 , :euclidean ) |&gt; Vettore . insert ( &quot;my_collection&quot; , % Vettore.Embedding { value : &quot;my_id&quot; , vector : [ 1.0 , 2.0 , 3.0 ] , metadata : %{ &quot;note&quot; =&gt; &quot;hello&quot; } } ) { :ok , &quot;my_id&quot; }","title":"Vettore.insert/3","ref":"Vettore.html#insert/3"},{"type":"function","doc":"Insert! one %Vettore.Embedding{} into the collection. Raise error if the vector is not numeric or the metadata is invalid. Otherwise, return the inserted value. Examples iex&gt; Vettore . new ( ) |&gt; Vettore . create_collection ( &quot;my_collection&quot; , 3 , :euclidean ) |&gt; Vettore . insert! ( &quot;my_collection&quot; , % Vettore.Embedding { value : &quot;my_id&quot; , vector : [ 1.0 , 2.0 , 3.0 ] , metadata : %{ &quot;note&quot; =&gt; %{ &quot;hello&quot; =&gt; &quot;world&quot; } } } ) ArgumentError [ … ]","title":"Vettore.insert!/3","ref":"Vettore.html#insert!/3"},{"type":"function","doc":"Allocate an empty in‑memory DB (owned by Rust). Keep the returned reference around – every other call expects it.","title":"Vettore.new/0","ref":"Vettore.html#new/0"},{"type":"function","doc":"Re‑rank an existing result list with Maximal Marginal Relevance . Options: :limit – desired output length (default 10 ) :alpha – relevance‑diversity balance 0.0..1.0 (default 0.5 ) # Examples iex&gt; Vettore . new ( ) |&gt; Vettore . create_collection ( &quot;my_collection&quot; , 3 , :euclidean ) |&gt; Vettore . insert ( &quot;my_collection&quot; , % Vettore.Embedding { value : &quot;my_id&quot; , vector : [ 1.0 , 2.0 , 3.0 ] , metadata : %{ &quot;note&quot; =&gt; &quot;hello&quot; } } ) |&gt; Vettore . insert ( &quot;my_collection&quot; , % Vettore.Embedding { value : &quot;my_id2&quot; , vector : [ 1.0 , 2.0 , 3.0 ] , metadata : %{ &quot;note&quot; =&gt; &quot;hello&quot; } } ) |&gt; Vettore . insert ( &quot;my_collection&quot; , % Vettore.Embedding { value : &quot;my_id3&quot; , vector : [ 1.0 , 2.0 , 3.0 ] , metadata : %{ &quot;note&quot; =&gt; &quot;hello&quot; } } ) |&gt; Vettore . rerank ( &quot;my_collection&quot; , [ { &quot;my_id&quot; , 0.0 } , { &quot;my_id2&quot; , 0.0 } , { &quot;my_id3&quot; , 0.0 } ] , limit : 1 ) { :ok , [ { &quot;my_id&quot; , 0.0 } ] }","title":"Vettore.rerank/4","ref":"Vettore.html#rerank/4"},{"type":"function","doc":"Similarity / nearest‑neighbour search. Options: :limit – number of results (default 10 ) :filter – metadata map; only embeddings whose metadata contains all key‑value pairs are considered. Examples iex&gt; Vettore . new ( ) |&gt; Vettore . create_collection ( &quot;my_collection&quot; , 3 , :euclidean ) |&gt; Vettore . insert ( &quot;my_collection&quot; , % Vettore.Embedding { value : &quot;my_id&quot; , vector : [ 1.0 , 2.0 , 3.0 ] , metadata : %{ &quot;note&quot; =&gt; &quot;hello&quot; } } ) |&gt; Vettore . similarity_search ( &quot;my_collection&quot; , [ 1.0 , 2.0 , 3.0 ] , limit : 1 ) { :ok , [ { &quot;my_id&quot; , 0.0 } ] }","title":"Vettore.similarity_search/4","ref":"Vettore.html#similarity_search/4"},{"type":"module","doc":"Stand-alone distance/similarity helpers and the collection-agnostic MMR re-ranker exposed by the Rust NIF layer.","title":"Vettore.Distance","ref":"Vettore.Distance.html"},{"type":"function","doc":"Compress a float vector into its sign-bit representation (64 floats → 64 bits → one u64 ). Useful for ultra-fast binary similarity. Examples iex&gt; Vettore.Distance . compress_f32_vector ( [ 1.0 , 2.0 , 3.0 ] ) [ 1 , 0 , 0 , ... .. 1 , 0 , 0 ]","title":"Vettore.Distance.compress_f32_vector/1","ref":"Vettore.Distance.html#compress_f32_vector/1"},{"type":"function","doc":"Cosine similarity in 0.0..1.0 ( (dot + 1) / 2 after length-normalisation). Examples iex&gt; Vettore.Distance . cosine ( [ 1 , 0 , 0 , 1 , 0 , 0 ] , [ 1 , 0 , 0 , 1 , 0 , 0 ] ) 0.0","title":"Vettore.Distance.cosine/2","ref":"Vettore.Distance.html#cosine/2"},{"type":"function","doc":"Raw dot product (no post-processing). Examples iex&gt; Vettore.Distance . dot_product ( [ 1 , 0 , 0 , 1 , 0 , 0 ] , [ 1 , 0 , 0 , 1 , 0 , 0 ] ) 1","title":"Vettore.Distance.dot_product/2","ref":"Vettore.Distance.html#dot_product/2"},{"type":"function","doc":"Similarity based on Euclidean (L2) distance. The result is in 0.0..1.0 via the mapping 1 / (1 + d) so that identical vectors yield 1.0 . Examples iex&gt; Vettore.Distance . euclidean ( [ 1 , 0 , 0 , 1 , 0 , 0 ] , [ 1 , 0 , 0 , 1 , 0 , 0 ] ) 0.0","title":"Vettore.Distance.euclidean/2","ref":"Vettore.Distance.html#euclidean/2"},{"type":"function","doc":"Bit-wise Hamming distance between two compressed vectors (see compress_f32_vector/1 ). Examples iex&gt; Vettore.Distance . hamming ( [ 1 , 0 , 0 , 1 , 0 , 0 ] , [ 1 , 0 , 0 , 1 , 0 , 0 ] ) 0","title":"Vettore.Distance.hamming/2","ref":"Vettore.Distance.html#hamming/2"},{"type":"function","doc":"MMR (Maximal-Marginal-Relevance) re-ranker that trades off query relevance and result diversity. initial – list of {id, similarity_to_query} tuples (first-pass hits) embeddings – {id, vector} pairs (dimension must be consistent) distance – &quot;euclidean&quot; | &quot;cosine&quot; | &quot;dot&quot; | &quot;binary&quot; alpha – 0 ⇢ only diversity, 1 ⇢ only query-relevance final_k – length of the wanted output list Examples iex&gt; Vettore.Distance . mmr_rerank ( [ { &quot;my_id&quot; , 0.0 } ] , [ { &quot;my_id&quot; , [ 1.0 , 2.0 , 3.0 ] } ] , &quot;euclidean&quot; , 0.5 , 1 ) [ { &quot;my_id&quot; , 0.0 } ]","title":"Vettore.Distance.mmr_rerank/5","ref":"Vettore.Distance.html#mmr_rerank/5"},{"type":"type","doc":"","title":"Vettore.Distance.alpha/0","ref":"Vettore.Distance.html#t:alpha/0"},{"type":"type","doc":"","title":"Vettore.Distance.distance/0","ref":"Vettore.Distance.html#t:distance/0"},{"type":"type","doc":"","title":"Vettore.Distance.embeddings/0","ref":"Vettore.Distance.html#t:embeddings/0"},{"type":"type","doc":"","title":"Vettore.Distance.final_k/0","ref":"Vettore.Distance.html#t:final_k/0"},{"type":"type","doc":"","title":"Vettore.Distance.search_result/0","ref":"Vettore.Distance.html#t:search_result/0"},{"type":"type","doc":"","title":"Vettore.Distance.vector/0","ref":"Vettore.Distance.html#t:vector/0"},{"type":"type","doc":"","title":"Vettore.Distance.vector_bits/0","ref":"Vettore.Distance.html#t:vector_bits/0"},{"type":"module","doc":"Represents a single embedding entry for insertion into a collection. Fields :value - A string or content identifier for this embedding can be Id, or Text (e.g. &quot;this is text data&quot;). :vector - A list of floating‑point numbers representing the embedding (e.g. [1.0, 2.0, 3.0] ). :metadata - (Optional) A map with any additional information you want to store (e.g. %{&quot;info&quot; =&gt; &quot;my note&quot;} ).","title":"Vettore.Embedding","ref":"Vettore.Embedding.html"},{"type":"type","doc":"","title":"Vettore.Embedding.t/0","ref":"Vettore.Embedding.html#t:t/0"},{"type":"extras","doc":"Vettore is a high-performance Elixir library for fast, in-memory operations on vector (embedding) data. It leverages a Rust backend via Rustler to store and manipulate vectors efficiently in a concurrent-safe HashMap .","title":"Vettore","ref":"readme.html"},{"type":"extras","doc":"Collections : Create named sets of embeddings with a fixed dimension and a choice of similarity metric. CRUD operations : Insert, batch-insert, retrieve, and delete embeddings by ID or by vector. Similarity search : Nearest-neighbor search with customizable :limit and optional metadata filtering. Reranking : Maximal Marginal Relevance (MMR) reranker for diversity-aware result reordering. Distance helpers : Standalone Euclidean, Cosine, Dot, and Hamming metrics, plus binary compression for ultra-fast comparisons.","title":"Vettore - Features","ref":"readme.html#features"},{"type":"extras","doc":"Add vettore to your list of dependencies in mix.exs : def deps do [ { :vettore , &quot;~&gt; 0.2.3&quot; } ] end Then fetch and compile: mix deps.get mix compile Note : The first compile will build the Rust crate; ensure you have a recent Rust toolchain installed.","title":"Vettore - Installation","ref":"readme.html#installation"},{"type":"extras","doc":"# 1. Start a new in-memory database reference db = Vettore . new ( ) # 2. Create a collection named &quot;my_collection&quot; with 3-dimensional vectors :ok = Vettore . create_collection ( db , &quot;my_collection&quot; , 3 , :euclidean ) # 3. Insert a single embedding embedding = % Vettore.Embedding { value : &quot;item_1&quot; , vector : [ 1.0 , 2.0 , 3.0 ] , metadata : %{ &quot;note&quot; =&gt; &quot;first vector&quot; } } :ok = Vettore . insert ( db , &quot;my_collection&quot; , embedding ) # 4. Retrieve by ID { :ok , emb } = Vettore . get_by_value ( db , &quot;my_collection&quot; , &quot;item_1&quot; ) IO . inspect ( emb . vector , label : &quot;Vector&quot; ) # 5. Similarity search (top 2 nearest neighbors) { :ok , results } = Vettore . similarity_search ( db , &quot;my_collection&quot; , [ 1.5 , 1.5 , 1.5 ] , limit : 2 ) IO . inspect ( results , label : &quot;Top-2 Results&quot; ) # 6. Rerank with MMR for diversity (alpha = 0.7) { :ok , reranked } = Vettore . rerank ( db , &quot;my_collection&quot; , results , limit : 2 , alpha : 0.7 ) IO . inspect ( reranked , label : &quot;MMR Reranked&quot; )","title":"Vettore - Quickstart","ref":"readme.html#quickstart"},{"type":"extras","doc":"Vettore.new/0 def new ( ) :: reference ( ) Allocates and returns an in-memory database handle backed by Rust. Vettore.create_collection/5 @spec create_collection ( db :: reference ( ) , name :: String . t ( ) , dim :: pos_integer ( ) , metric :: :euclidean | :cosine | :dot | :hnsw | :binary , opts :: [ keep_embeddings : boolean ( ) ] ) :: { :ok , String . t ( ) } | { :error , String . t ( ) } name : Collection identifier dim : Dimensionality of vectors metric : Similarity measure opts : :keep_embeddings (default: true ) — whether to retain embeddings on deletion Vettore.insert/3 @spec insert ( db :: reference ( ) , collection :: String . t ( ) , embedding :: Vettore.Embedding . t ( ) ) :: { :ok , String . t ( ) } | { :error , String . t ( ) } Insert a single %Vettore.Embedding{} struct into the named collection. Vettore.batch/3 @spec batch ( db :: reference ( ) , collection :: String . t ( ) , embeddings :: [ Vettore.Embedding . t ( ) ] ) :: { :ok , [ String . t ( ) ] } | { :error , String . t ( ) } Batch-insert multiple embeddings at once; non-embedding elements are ignored. Retrieval and Deletion Vettore.get_by_value/3 — fetch by embedding ID Vettore.get_by_vector/3 — fetch by exactly matching vector Vettore.get_all/2 — returns all {value, vector, metadata} Vettore.delete/3 — delete by ID Vettore.similarity_search/4 @spec similarity_search ( db :: reference ( ) , collection :: String . t ( ) , query :: [ number ( ) ] , opts :: [ limit : pos_integer ( ) , filter : map ( ) ] ) :: { :ok , [ { String . t ( ) , float ( ) } ] } | { :error , String . t ( ) } limit (default: 10 ) filter : metadata map to pre-filter embeddings Vettore.rerank/4 (MMR) @spec rerank ( db :: reference ( ) , collection :: String . t ( ) , initial :: [ { String . t ( ) , number ( ) } ] , opts :: [ limit : pos_integer ( ) , alpha : float ( ) ] ) :: { :ok , [ { String . t ( ) , number ( ) } ] } | { :error , String . t ( ) } alpha : 0.0..1.0 balance between relevance and diversity","title":"Vettore - API Reference","ref":"readme.html#api-reference"},{"type":"extras","doc":"You can call these functions without creating a DB or collection: Vettore.Distance . euclidean ( [ 1.0 , 2.0 ] , [ 2.0 , 3.0 ] ) # =&gt; 1 / (1 + L2) Vettore.Distance . cosine ( [ 1 , 0 ] , [ 0 , 1 ] ) # =&gt; (dot + 1) / 2 Vettore.Distance . dot_product ( [ 1 , 2 ] , [ 3 , 4 ] ) # =&gt; raw dot Vettore.Distance . hamming ( bits1 , bits2 ) # =&gt; Hamming distance bits = Vettore.Distance . compress_f32_vector ( [ 0.1 , 0.4 ] ) # MMR re-ranker standalone (collection-agnostic) initial = [ { &quot;id1&quot; , 0.9 } , { &quot;id2&quot; , 0.85 } , ... ] embeds = [ { &quot;id1&quot; , [ v1 ... ] } , { &quot;id2&quot; , [ v2 ... ] } , ... ] Vettore.Distance . mmr_rerank ( initial , embeds , &quot;cosine&quot; , 0.5 , 5 ) Example For example you can create Vector DB with ETS backend: defmodule ETSVectorDB do @moduledoc &quot;&quot;&quot; A simple ETS-backed vector store with k-NN search using cosine similarity. &quot;&quot;&quot; alias Vettore.Distance , as : Dist @type id :: String . t ( ) | atom ( ) @type vector :: [ float ( ) ] @type table :: atom ( ) @doc &quot;&quot;&quot; Creates (or re-creates) an ETS table to hold your vectors. &quot;&quot;&quot; @spec create_table ( table ( ) ) :: :ok def create_table ( table \\\\ :vector_db ) do # If it already exists, delete and recreate if :ets . info ( table ) != :undefined , do : :ets . delete ( table ) :ets . new ( table , [ :set , :public , read_concurrency : true ] ) :ok end @doc &quot;&quot;&quot; Inserts or updates a vector under the given id. &quot;&quot;&quot; @spec put ( table ( ) , id ( ) , vector ( ) ) :: :ok def put ( table \\\\ :vector_db , id , vector ) when is_list ( vector ) do :ets . insert ( table , { id , vector } ) :ok end @doc &quot;&quot;&quot; Finds the top `k` vectors most similar to `query_vec` using cosine similarity. Returns a list of `{id, score}` sorted by descending score (highest = most similar). &quot;&quot;&quot; @spec search ( table ( ) , vector ( ) , pos_integer ( ) , keyword ( ) ) :: [ { id ( ) , float ( ) } ] def search ( table \\\\ :vector_db , query_vec , k , opts \\\\ [ ] ) when is_list ( query_vec ) and is_integer ( k ) and k &gt; 0 do # you could add other options here, e.g. distance: :euclidean :ets . foldl ( fn { id , vec } , acc -&gt; score = Dist . cosine ( query_vec , vec ) [ { id , score } | acc ] end , [ ] , table ) |&gt; Enum . sort_by ( fn { _id , score } -&gt; - score end ) |&gt; Enum . take ( k ) end end","title":"Vettore - Distance Helpers (Vettore.Distance)","ref":"readme.html#distance-helpers-vettore-distance"},{"type":"extras","doc":"The similarity_search function works as follows: It retrieves the target collection and verifies that the query vector’s dimension matches. Depending on the chosen distance metric, it selects an appropriate function: Euclidean: Computes the standard Euclidean distance. Cosine / DotProduct: Computes the dot product (with normalization applied for Cosine). HNSW: Uses a graph-based approach for approximate nearest neighbor search. Binary: Compresses the query vector into a binary signature and computes the Hamming distance between this signature and those of all stored embeddings. For every embedding in the collection, it calculates a “score” between the stored vector (or its compressed representation) and the query. The results are sorted: For Euclidean distance , lower scores (closer to zero) are better. For Cosine/DotProduct , higher scores are considered more similar. For Binary , a lower Hamming distance means the vectors are more similar. Finally, the top‑k results are returned as a list of tuples (embedding_id, score) . Technique/Algorithm Measures Magnitude Sensitive?¹ Scale Invariant?¹ Best Use Cases Pros Cons Euclidean Distance Straight-line distance Yes No Dense data where both magnitude &amp; direction are important Intuitive, widely used, captures magnitude differences Sensitive to scale differences, high dimensionality issues Cosine Similarity Directional similarity (angle) No Yes Text or high-dimensional data where scale invariance is desired Insensitive to magnitude, works well with normalized vectors Ignores magnitude differences Dot Product Combination of direction &amp; magnitude Yes No Applications where both direction &amp; magnitude matter Computationally efficient, captures both aspects Sensitive to vector magnitudes HNSW Indexing Graph-based Approximate Nearest Neighbor Search Dependent on Metric Dependent on Metric Large datasets, real-time search when approximate results are acceptable Sublinear search time , good speed-accuracy trade-off, scalable Approximate results, index build time and memory overhead Binary (Hamming) Fast binary signature comparison using Hamming distance No Yes Applications requiring ultra‑fast approximate searches on large-scale data Extremely fast comparison via bit-level operations, low memory footprint Loses precision due to compression, less suited when exact distances are needed","title":"Vettore - Similarity Search","ref":"readme.html#similarity-search"},{"type":"extras","doc":"HNSW can speed up searches significantly for large datasets but comes with higher memory usage for the index. Binary distance uses bit-level compression and Hamming distance for extremely fast approximate similarity checks (especially beneficial for large or high-dimensional vectors, though it trades off some precision). Cosine normalizes vectors once on insertion, so queries and stored embeddings use a straightforward dot product. Dot Product directly multiplies corresponding elements. Euclidean uses a SIMD approach ( wide::f32x4 ) for partial vectorization.","title":"Vettore - Performance Notes","ref":"readme.html#performance-notes"},{"type":"extras","doc":"Contributions are welcome! Please open an issue or submit a PR. Fork the repo Create a feature branch Add tests in test/ Submit a PR","title":"Vettore - Contributing","ref":"readme.html#contributing"},{"type":"extras","doc":"Apache 2.0 LICENSE","title":"Vettore - License","ref":"readme.html#license"},{"type":"extras","doc":"Apache License Version 2.0, January 2004 http://www.apache.org/licenses/ TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION 1. Definitions. &quot;License&quot; shall mean the terms and conditions for use, reproduction, and distribution as defined by Sections 1 through 9 of this document. &quot;Licensor&quot; shall mean the copyright owner or entity authorized by the copyright owner that is granting the License. &quot;Legal Entity&quot; shall mean the union of the acting entity and all other entities that control, are controlled by, or are under common control with that entity. For the purposes of this definition, &quot;control&quot; means (i) the power, direct or indirect, to cause the direction or management of such entity, whether by contract or otherwise, or (ii) ownership of fifty percent (50%) or more of the outstanding shares, or (iii) beneficial ownership of such entity. &quot;You&quot; (or &quot;Your&quot;) shall mean an individual or Legal Entity exercising permissions granted by this License. &quot;Source&quot; form shall mean the preferred form for making modifications, including but not limited to software source code, documentation source, and configuration files. &quot;Object&quot; form shall mean any form resulting from mechanical transformation or translation of a Source form, including but not limited to compiled object code, generated documentation, and conversions to other media types. &quot;Work&quot; shall mean the work of authorship, whether in Source or Object form, made available under the License, as indicated by a copyright notice that is included in or attached to the work (an example is provided in the Appendix below). &quot;Derivative Works&quot; shall mean any work, whether in Source or Object form, that is based on (or derived from) the Work and for which the editorial revisions, annotations, elaborations, or other modifications represent, as a whole, an original work of authorship. For the purposes of this License, Derivative Works shall not include works that remain separable from, or merely link (or bind by name) to the interfaces of, the Work and Derivative Works thereof. &quot;Contribution&quot; shall mean any work of authorship, including the original version of the Work and any modifications or additions to that Work or Derivative Works thereof, that is intentionally submitted to Licensor for inclusion in the Work by the copyright owner or by an individual or Legal Entity authorized to submit on behalf of the copyright owner. For the purposes of this definition, &quot;submitted&quot; means any form of electronic, verbal, or written communication sent to the Licensor or its representatives, including but not limited to communication on electronic mailing lists, source code control systems, and issue tracking systems that are managed by, or on behalf of, the Licensor for the purpose of discussing and improving the Work, but excluding communication that is conspicuously marked or otherwise designated in writing by the copyright owner as &quot;Not a Contribution.&quot; &quot;Contributor&quot; shall mean Licensor and any individual or Legal Entity on behalf of whom a Contribution has been received by Licensor and subsequently incorporated within the Work. 2. Grant of Copyright License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable copyright license to reproduce, prepare Derivative Works of, publicly display, publicly perform, sublicense, and distribute the Work and such Derivative Works in Source or Object form. 3. Grant of Patent License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable (except as stated in this section) patent license to make, have made, use, offer to sell, sell, import, and otherwise transfer the Work, where such license applies only to those patent claims licensable by such Contributor that are necessarily infringed by their Contribution(s) alone or by combination of their Contribution(s) with the Work to which such Contribution(s) was submitted. If You institute patent litigation against any entity (including a cross-claim or counterclaim in a lawsuit) alleging that the Work or a Contribution incorporated within the Work constitutes direct or contributory patent infringement, then any patent licenses granted to You under this License for that Work shall terminate as of the date such litigation is filed. 4. Redistribution. You may reproduce and distribute copies of the Work or Derivative Works thereof in any medium, with or without modifications, and in Source or Object form, provided that You meet the following conditions: (a) You must give any other recipients of the Work or Derivative Works a copy of this License; and (b) You must cause any modified files to carry prominent notices stating that You changed the files; and (c) You must retain, in the Source form of any Derivative Works that You distribute, all copyright, patent, trademark, and attribution notices from the Source form of the Work, excluding those notices that do not pertain to any part of the Derivative Works; and (d) If the Work includes a &quot;NOTICE&quot; text file as part of its distribution, then any Derivative Works that You distribute must include a readable copy of the attribution notices contained within such NOTICE file, excluding those notices that do not pertain to any part of the Derivative Works, in at least one of the following places: within a NOTICE text file distributed as part of the Derivative Works; within the Source form or documentation, if provided along with the Derivative Works; or, within a display generated by the Derivative Works, if and wherever such third-party notices normally appear. The contents of the NOTICE file are for informational purposes only and do not modify the License. You may add Your own attribution notices within Derivative Works that You distribute, alongside or as an addendum to the NOTICE text from the Work, provided that such additional attribution notices cannot be construed as modifying the License. You may add Your own copyright statement to Your modifications and may provide additional or different license terms and conditions for use, reproduction, or distribution of Your modifications, or for any such Derivative Works as a whole, provided Your use, reproduction, and distribution of the Work otherwise complies with the conditions stated in this License. 5. Submission of Contributions. Unless You explicitly state otherwise, any Contribution intentionally submitted for inclusion in the Work by You to the Licensor shall be under the terms and conditions of this License, without any additional terms or conditions. Notwithstanding the above, nothing herein shall supersede or modify the terms of any separate license agreement you may have executed with Licensor regarding such Contributions. 6. Trademarks. This License does not grant permission to use the trade names, trademarks, service marks, or product names of the Licensor, except as required for reasonable and customary use in describing the origin of the Work and reproducing the content of the NOTICE file. 7. Disclaimer of Warranty. Unless required by applicable law or agreed to in writing, Licensor provides the Work (and each Contributor provides its Contributions) on an &quot;AS IS&quot; BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied, including, without limitation, any warranties or conditions of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A PARTICULAR PURPOSE. You are solely responsible for determining the appropriateness of using or redistributing the Work and assume any risks associated with Your exercise of permissions under this License. 8. Limitation of Liability. In no event and under no legal theory, whether in tort (including negligence), contract, or otherwise, unless required by applicable law (such as deliberate and grossly negligent acts) or agreed to in writing, shall any Contributor be liable to You for damages, including any direct, indirect, special, incidental, or consequential damages of any character arising as a result of this License or out of the use or inability to use the Work (including but not limited to damages for loss of goodwill, work stoppage, computer failure or malfunction, or any and all other commercial damages or losses), even if such Contributor has been advised of the possibility of such damages. 9. Accepting Warranty or Additional Liability. While redistributing the Work or Derivative Works thereof, You may choose to offer, and charge a fee for, acceptance of support, warranty, indemnity, or other liability obligations and/or rights consistent with this License. However, in accepting such obligations, You may act only on Your own behalf and on Your sole responsibility, not on behalf of any other Contributor, and only if You agree to indemnify, defend, and hold each Contributor harmless for any liability incurred by, or claims asserted against, such Contributor by reason of your accepting any such warranty or additional liability. END OF TERMS AND CONDITIONS APPENDIX: How to apply the Apache License to your work. To apply the Apache License to your work, attach the following boilerplate notice, with the fields enclosed by brackets &quot;[]&quot; replaced with your own identifying information. (Don't include the brackets!) The text should be enclosed in the appropriate comment syntax for the file format. We also recommend that a file or class name and description of purpose be included on the same &quot;printed page&quot; as the copyright notice for easier identification within third-party archives. Copyright [yyyy] [name of copyright owner] Licensed under the Apache License, Version 2.0 (the &quot;License&quot;); you may not use this file except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an &quot;AS IS&quot; BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.","title":"LICENSE","ref":"license.html"}]